# from http://stackoverflow.com/questions/3841725/how-to-launch-tests-for-django-reusable-app
import sys
from django.conf import settings

settings.configure(DEBUG=True,
               DATABASES={
                    'default': {
                        'ENGINE': 'django.db.backends.sqlite3',
                    }
                },
               ROOT_URLCONF='timelinejs.urls',
               INSTALLED_APPS=('django.contrib.auth',
                              'django.contrib.contenttypes',
                              'django.contrib.sessions',
                              'django.contrib.admin',
                              'timelinejs',))

from django.test.simple import DjangoTestSuiteRunner
test_runner = DjangoTestSuiteRunner(verbosity=1)
failures = test_runner.run_tests(['timelinejs', ])
if failures:
    sys.exit(failures)



# Create your views here.
from django.views.generic import DetailView, ListView, FormView
from django.utils.decorators import method_decorator
from django.contrib.auth.decorators import permission_required
from django.contrib import messages
from django.template.defaultfilters import slugify
from django.db import transaction

from .models import Timeline, TimelineImporter
from .forms import TimelineForm

class TimelineListView(ListView):
    """List timelines."""
    context_object_name='timelines'
    template_name = 'timelinejs/timelines.html'

    def get_queryset(self):
        return Timeline.objects.visible_to_user(self.request.user)

    def get_context_data(self, **kwargs):
        context = super(TimelineListView, self).get_context_data(**kwargs)
        context['user_can_add_timelines'] = False
        if self.request.user.is_staff and self.request.user.has_perm('timelinejs.add_timeline'):
            context['user_can_add_timelines'] = True
        return context

class TimelineDetailView(DetailView):
    """Show a timeline."""
    template_name = 'timelinejs/timeline.html'

    def get_object(self):
        return Timeline.objects.get_visible_to_user_or_404(
                user=self.request.user,
                slug=self.kwargs['slug'])

class ImportTimelineFromSpreadsheetView(FormView):
    """Import a timeline from a Google spreadsheet."""
    form_class = TimelineForm
    template_name = 'timelinejs/import_timeline.html'

    @method_decorator(permission_required('timelinejs.add_timeline'))
    def dispatch(self, *args, **kwargs):
        return super(ImportTimelineFromSpreadsheetView, self).dispatch(*args, **kwargs)

    def form_valid(self, form):
        with transaction.commit_manually():
            timeline = form.save(commit=False)
            timeline.slug = slugify(timeline.title)
            timeline.save()
            self.timeline = timeline
            try:
                importer = TimelineImporter(form.cleaned_data['item_data'], timeline)
                recoverable_errors = importer.import_items()
            except Exception, e:
                messages.add_message(self.request, messages.ERROR, 'Importing your timeline data failed with error: %s. You may be able to fix this by editing the spreadsheet you are trying to import.' % e)
                transaction.rollback()
                return self.form_invalid(form)
            # simply warn user of recoverable errors in the spreadsheet that were ignored
            for error in recoverable_errors:
                messages.add_message(self.request, messages.WARNING, error)
            transaction.commit()
        return super(ImportTimelineFromSpreadsheetView, self).form_valid(form)

    def get_success_url(self):
        return self.timeline.get_absolute_url()



from django import forms

from .models import Timeline

class TimelineForm(forms.ModelForm):
    item_data = forms.CharField(widget=forms.HiddenInput)
    class Meta:
        model = Timeline
        exclude = ('slug',)



import re

MAX_URL_LENGTH = 35
TRAILING_URL_LENGTH = 6
SHORTENED_URL_PLACEHOLDER = "..."

def get_link_matches(text):
    re_text = r'(?i)\b(?<!..]\(|\d]:\s)((?:https?://|(?<!:\/\/)www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:\'".,<>?]))'
    matcher = re.compile(re_text)
    return matcher.findall(text)

def markdown_linkify(text):
    """
    Preprocessor to add links to markdown.
    """
    for match in get_link_matches(text):
        url = match[0]
        # add http:// in front of links without it
        if url[:4] != "http":
            href = "http://" + url
        else:
            href = url
        # shorten the link_text if necessary
        if len(url) > MAX_URL_LENGTH:
            link_text = url[:MAX_URL_LENGTH - (TRAILING_URL_LENGTH + len(SHORTENED_URL_PLACEHOLDER))] + SHORTENED_URL_PLACEHOLDER + url[-TRAILING_URL_LENGTH:]
        else:
            link_text = url
        text = text.replace(url, "[%s](%s)" % (link_text, href))
    return text



import markdown, datetime
from django.db import models
from django.utils import simplejson as json, html
from django.conf import settings
from django.shortcuts import get_object_or_404
from django.core.urlresolvers import reverse

from .utils import markdown_linkify

# Create your models here.

TITLE = 't'
ERA = 'e'
NORMAL = 'n'
ITEM_TYPE_CHOICES = (
        (TITLE, 'Title',),
        (ERA, 'Era',),
        (NORMAL, 'Normal',),
)
VIEW_PRIVATE_TIMELINES_PERM = 'timelinejs.view_private_timelines'

data_source_help_text = "Leave blank unless you are using a Google Spreadsheet."
class TimelineManager(models.Manager):
    def visible_to_user(self, user=None):
        filter_kwargs = {'published': True}
        if user is None or not user.has_perm(VIEW_PRIVATE_TIMELINES_PERM):
            filter_kwargs['private'] = False
        return self.filter(**filter_kwargs)

    def get_visible_to_user_or_404(self, user=None, **kwargs):
        kwargs['published'] = True
        if user is None or not user.has_perm(VIEW_PRIVATE_TIMELINES_PERM):
            kwargs['private'] = False
        return get_object_or_404(Timeline, **kwargs)

class Timeline(models.Model):
    data_source = models.URLField(max_length=255, null=True, blank=True, help_text=data_source_help_text)
    title = models.CharField(max_length=255, unique=True)
    slug = models.SlugField(unique=True)
    private = models.BooleanField(default=False)
    published = models.BooleanField(default=True)

    objects = TimelineManager()

    class Meta:
        permissions = (
                (VIEW_PRIVATE_TIMELINES_PERM.split('.')[1], 'Can see published, private timelines.'),
        )

    def get_absolute_url(self):
        return reverse('timeline', kwargs={'slug': self.slug})

    def __unicode__(self):
        return self.title

    @property
    def source(self):
        if self.timelineitem_set.all().count():
            return self.process_to_json()
        else:
            return '"%s"' % self.data_source

    def process_to_json(self):
        parser = TimelineParser(self.timelineitem_set.all())
        parser.parse()
        return parser.get_json()

    def load_items_from_json(self, item_json):
        importer = TimelineImporter(item_json, self)
        errors = importer.import_items()
        return errors



item_type_help_text = "You must have one and only one title slide. You may use multiple era slides."
class TimelineItem(models.Model):
    timeline = models.ForeignKey(Timeline, blank=True, null=True)
    start_date = models.DateField()
    end_date = models.DateField(null=True, blank=True)
    item_type = models.CharField(max_length=1, choices=ITEM_TYPE_CHOICES, default=NORMAL, help_text=item_type_help_text)
    headline = models.TextField(default="", blank=True)
    _headline = models.TextField(default="", blank=True, help_text="Processed headline")
    text = models.TextField(default="", blank=True, help_text="Feel free to use Markdown syntax. Your title slide **must** have text in order to show up.")
    _text = models.TextField(default="", blank=True, help_text="Processesed text.")
    tag = models.CharField(max_length=255, null=True, blank=True, help_text="You may use up to six tags per timeline.")
    _tag = models.CharField(max_length=255, null=True, blank=True, help_text="Processed tag.")
    media = models.CharField(max_length=255, null=True, blank=True, help_text="Enter an external link. If you are using the Media Upload field this will be filled in automatically.")
    _media = models.CharField(max_length=255, null=True, blank=True, help_text="Processed media.")
    media_upload = models.FileField(upload_to="timelinemedia", max_length=250, blank=True, null=True)
    media_caption = models.CharField(max_length=255, null=True, blank=True)
    _media_caption = models.CharField(max_length=255, null=True, blank=True, help_text="Processed caption.")
    media_credit = models.CharField(max_length=255, null=True, blank=True)
    _media_credit = models.CharField(max_length=255, null=True, blank=True, help_text="Processed credit.")
    media_thumbnail = models.CharField(max_length=255, null=True, blank=True, help_text="Must be a link to an image no bigger than 32x32 pixels.")
    _media_thumbnail = models.CharField(max_length=255, null=True, blank=True, help_text="Processed thumbnail.")

    def _process_date(self, d):
        return "%d,%d,%d" % (d.year, d.month, d.day)

    def process_to_json_dict(self):
        json_dict = {'startDate': self._process_date(self.start_date),}
        if self.end_date:
            json_dict['endDate'] = self._process_date(self.end_date)
        if self.headline:
            json_dict['headline'] = self._headline
        if self.text:
            json_dict['text'] = self._text
        if self.text:
            json_dict['tag'] = self._tag
        if self.media:
            asset = {'media': self._media}
            if self.media_caption:
                asset['caption'] = self._media_caption
            if self.media_credit:
                asset['credit'] = self._media_credit
            if self.media_thumbnail:
                asset['thumbnail'] = self._media_thumbnail
            json_dict['asset'] = asset
        return json_dict

    def save(self, *args, **kwargs):
        # process text attributes
        self.process_fields()
        # need to super first so we generate the real filename to get the true
        # url for the media
        super(TimelineItem, self).save(*args, **kwargs)
        self._set_media_from_media_upload(*args, **kwargs)

    def process_fields(self):
        # process the text attributes
        escaped_attrs = ['headline', 'text', 'tag', 'media', 'media_caption', 'media_credit', 'media_thumbnail']
        for attr in escaped_attrs:
            attr_val = getattr(self, attr, None)
            if attr_val:
                setattr(self, '_' + attr, html.escape(attr_val))
        if self._text:
            self._text = markdown.markdown(markdown_linkify(self._text))

    def _set_media_from_media_upload(self, *args, **kwargs):
        if self.media_upload and self._media != html.escape(self.media_upload.url):
            self.media = "%s%s" % (settings.FULLY_QUALIFIED_DOMAIN, self.media_upload.url)
            self._media = html.escape(self.media)
            super(TimelineItem, self).save(*args, **kwargs)

    def load_from_object(self, obj):
        obj.load_into_timeline_item(self)
        self.process_fields()


class TimelineParser(object):
    """
    Takes an iterable of TimelineItems and returns parsed json.
    """
    def __init__(self, items):
        self.items = items
        self.processed_items = None
        self.timeline = None
        self.has_title_item = False

    def parse(self):
        # set up root
        self.timeline = {'timeline': 
                {
                    'type': 'default',
                    'date': [],
                    'era': [],
                },
        }
        for item in self.items:
            self.add_item(item)

    def add_item(self, item):
        root = self.timeline['timeline']
        processed = item.process_to_json_dict()
        if item.item_type == TITLE:
            # skip extra title items
            if self.has_title_item:
                root['date'].append(processed)
            else:
                root.update(processed)
                self.has_title_item = True
        elif item.item_type == ERA:
            root['era'].append(processed)
        elif item.item_type == NORMAL:
            root['date'].append(processed)

    def get_json(self):
        if self.timeline is None:
            self.parse()
        return json.dumps(self.timeline)

class TimelineImporter(object):

    def __init__(self, item_json, timeline=None):
        self.item_json = item_json
        self.timeline = timeline


    def _parse_timeline_date(self, d):
        try:
            split = [int(i) for i in d.split()[0].split('/')]
            if len(split) != 3:
                return None
            return datetime.date(int(split[2]), int(split[0]), int(split[1]))
        except:
            pass

    def import_items(self):
        errors = []
        items = json.loads(self.item_json)
        root = items['timeline']
        # create title slide
        title_props = {'item_type': TITLE, 'timeline': self.timeline}
        title_props.update(self._get_basic_props(root))
        title_props.update(self._get_media_props(root))
        TimelineItem.objects.create(**title_props)

        # create era slides
        for era_slide in root.get('era', []):
            try:
                era_props =  {'item_type': ERA, 'timeline': self.timeline}
                era_props.update(self._get_basic_props(era_slide))
                TimelineItem.objects.create(**era_props)
            except Exception, e:
                errors.append('Failed to parse Era entry with headline "%s" so it was omitted. Error was %s.' % (era_slide.get('headline', '<No Headline>'), e))

        # create normal slides
        for normal_slide in root.get('date', []):
            try:
                normal_props =  {'item_type': NORMAL, 'timeline': self.timeline}
                normal_props.update(self._get_basic_props(normal_slide))
                normal_props.update(self._get_media_props(normal_slide))
                TimelineItem.objects.create(**normal_props)
            except Exception, e:
                errors.append('Failed to parse entry with headline "%s" so it was omitted. Error was %s.' % (normal_slide.get('headline', '<No Headline>'), e))
        return errors

    def _get_basic_props(self, obj):
        props = {
                'start_date': self._parse_timeline_date(obj['startDate']),
        }
        props['end_date'] = self._parse_timeline_date(obj.get('endDate', None))
        props['headline'] = obj.get('headline', '')
        props['text'] = obj.get('text', '')
        props['tag'] = obj.get('tag', None)
        return props

    def _get_media_props(self, obj):
        props = {}
        if obj.has_key('asset'):
            asset = obj['asset']
            # should throw an error if we don't have a media value
            props['media'] = asset['media']
            props['media_caption'] = asset.get('caption', None)
            props['media_credit'] = asset.get('credit', None)
            props['media_thumbnail'] = asset.get('thumbnail', None)
        return props



from django.conf.urls import patterns, url, include
from django.contrib import admin
admin.autodiscover()

from .views import TimelineDetailView, TimelineListView, ImportTimelineFromSpreadsheetView

urlpatterns = patterns('',
    url(r'^$', TimelineListView.as_view(), name='timelines'),
    url(r'^import/$', ImportTimelineFromSpreadsheetView.as_view(), name='import_timeline_from_spreadsheet'),
    # this is for tests, you probably want to remove this or use your own urls.py
    url(r'^admin/', include(admin.site.urls)),
    url(r'^(?P<slug>[a-zA-Z0-9-_]+)/$', TimelineDetailView.as_view(), name='timeline'),
)



"""
This file demonstrates writing tests using the unittest module. These will pass
when you run "manage.py test".

Replace this with more appropriate tests for your application.
"""
import os

from django.test import TestCase
from django.test.client import Client
from django.http import Http404
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User, Permission

from .models import Timeline, TimelineImporter


class TestTimelinePermissions(TestCase):

    def setUp(self):
        self.timeline = Timeline.objects.create(
                title='Test Timeline',
                slug='test-timeline',
        )
        self.c = Client()
        self.u = User.objects.create_user('test', password='test')
        self.u.is_staff = True
        self.u.save()
        p1 = Permission.objects.get_by_natural_key('view_private_timelines', 'timelinejs', 'timeline')
        p2 = Permission.objects.get_by_natural_key('add_timeline', 'timelinejs', 'timeline')
        self.u.user_permissions.add(p1)
        self.u.user_permissions.add(p2)

    def test_private_timelines_anonymous(self):
        """
        Test visibility of private timelines.
        """
        ts = Timeline.objects.visible_to_user()
        self.assertEquals(1, len(ts), "Anonymous user should see public timelines.")
        self.timeline.private = True
        self.timeline.save()
        ts = Timeline.objects.visible_to_user()
        self.assertEquals(0, len(ts), "Anonymous user should not see private timelines.")

    def test_private_timelines_anonymous_ui(self):
        """
        Test visibility of private timelines.
        """
        resp = self.c.get(reverse('timelines'))
        self.assertContains(resp, self.timeline.title)
        self.timeline.private = True
        self.timeline.save()
        resp = self.c.get(reverse('timelines'))
        self.assertNotContains(resp, self.timeline.title)

    def test_add_timelines_anonymous_ui(self):
        resp = self.c.get(reverse('timelines'))
        self.assertNotContains(resp, reverse('admin:timelinejs_timeline_add'))

    def test_add_timelines_authenticated_ui(self):
        self.c.login(username='test', password='test')
        resp = self.c.get(reverse('timelines'))
        self.assertContains(resp, reverse('admin:timelinejs_timeline_add'))

    def test_private_timelines_authenticated(self):
        self.timeline.private = True
        self.timeline.save()
        ts = Timeline.objects.visible_to_user(user=self.u)
        self.assertEquals(1, len(ts), "Authenticated user should see private timelines.")

    def test_private_timelines_authenticated_ui(self):
        """
        Test visibility of private timelines.
        """
        self.c.login(username='test', password='test')
        resp = self.c.get(reverse('timelines'))
        self.assertContains(resp, self.timeline.title)
        self.timeline.private = True
        self.timeline.save()
        resp = self.c.get(reverse('timelines'))
        self.assertContains(resp, self.timeline.title)

    def test_published_timelines(self):
        """
        Test visibility of published timelines.
        """
        ts = Timeline.objects.visible_to_user()
        self.assertEquals(1, len(ts), "Anonymous user should see public published timelines.")
        self.timeline.published = False
        self.timeline.save()
        ts = Timeline.objects.visible_to_user()
        self.assertEquals(0, len(ts), "Anonymous user should not see unpublished timelines.")
        ts = Timeline.objects.visible_to_user(user=self.u)
        self.assertEquals(0, len(ts), "Authenticated user should not see unpublished timelines.")

    def test_private_timeline_anon(self):
        # test public timeline with anonymous
        Timeline.objects.get_visible_to_user_or_404(slug=self.timeline.slug)
        # set to private
        self.timeline.private = True
        self.timeline.save()
        with self.assertRaises(Http404):
            Timeline.objects.get_visible_to_user_or_404(slug=self.timeline.slug)

    def test_private_timeline_auth(self):
        self.timeline.private = True
        self.timeline.save()
        Timeline.objects.get_visible_to_user_or_404(self.u, slug=self.timeline.slug)

    def test_unpublished_timeline_anon(self):
        # set to unpublished
        self.timeline.published = False
        self.timeline.save()
        with self.assertRaises(Http404):
            Timeline.objects.get_visible_to_user_or_404(slug=self.timeline.slug)

    def test_unpublished_timeline_auth(self):
        # set to unpublished
        self.timeline.published = False
        self.timeline.save()
        with self.assertRaises(Http404):
            Timeline.objects.get_visible_to_user_or_404(self.u, slug=self.timeline.slug)

    def test_unpublished_timeline_anon_ui(self):
        # set to unpublished
        self.timeline.published = False
        self.timeline.save()
        resp = self.c.get(reverse('timeline', args=[self.timeline.slug]))
        self.assertEquals(404, resp.status_code)

    def test_unpublished_timeline_auth_ui(self):
        # set to unpublished
        self.timeline.published = False
        self.timeline.save()
        self.c.login(username='test', password='test')
        resp = self.c.get(reverse('timeline', args=[self.timeline.slug]))
        self.assertEquals(404, resp.status_code)

class TestTimelineImport(TestCase):
    def setUp(self):
        self.timeline = Timeline.objects.create(
                title='Test Timeline',
                slug='test-timeline',
        )

    def test_timeline_import(self):
        timeline_json = open(os.path.join(os.path.dirname(__file__), 'fixtures/pastor_john_timeline.json')).readline()
        timeline_json.strip()
        importer = TimelineImporter(timeline_json, self.timeline)
        errors = importer.import_items()
        self.assertEquals(30, self.timeline.timelineitem_set.count())
        self.assertEquals(0, len(errors))



from django.contrib import admin
from .models import Timeline, TimelineItem

class TimelineItemInline(admin.StackedInline):
    model = TimelineItem
    extra = 3
    exclude = (
            '_headline',
            '_text',
            '_media',
            '_media_caption',
            '_media_credit',
            '_media_thumbnail',
            '_tag',
            )

class TimelineAdmin(admin.ModelAdmin):
    prepopulated_fields = {'slug': ('title',)}
    inlines = [
            TimelineItemInline,
            ]


admin.site.register(Timeline, TimelineAdmin)



# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
    name='dj-timelinejs',
    version='0.1.1',
    author='Ben Best',
    author_email='ben.best@gmail.com,',
    packages=['timelinejs'],
    use_package_data=True,
    url='https://github.com/azundo/dj-timelinejs/',
    license='BSD license and MPL 2.0, see LICENSE',
    description='Support for serving TimelineJS from Django sites.',
    long_description=open('README.md').read(),
)

